
/* darkstat: a network traffic analyzer
 * (c) 2001, Emil Mikulic.
 */

/* Main body program */

/* If any of this looks whacked it's because I code in a 100x50
 * SSH terminal window.
 *
 * Everything here is covered by the GPL (see the 'COPYING' file)
 */

#include <pcap.h>
#include <netinet/in.h> /* for ntohs() and friends */
#include <net/if.h>
#include <arpa/inet.h>

#include <assert.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include "version.h"
#include "host_db.h"
#include "proto_db.h"
#include "port_db.h"
#include "bignum.h"
#include "dns.h"
#include "www.h"
#include "graph.h"

#define DBFILE "darkstat.db"

#define SAVE_EVERY_x_PACKETS 200

i64 num_packets, total_data;

dword local_ip = 0;
int want_packetdump = 1, is_ether = 1;
time_t t_start, t_already = 0;



void save_db(char *filename);

void freakout(void)
{
	fprintf(stderr, "\nSomething has gone terribly wrong.\n\
I'm going to try to dump db to minimize data loss.\n");

	save_db("darkstat.err");
	fprintf(stderr, "Dying now...\n");
	exit(101);
}



void print_64bit(const i64 val)
{
	char *buf = bignum_from_64bit(val);
	printf(buf);
	free(buf);
}



void print_binary(const byte x)
{
	putchar(x & (1<<7) ? '1' : '0');
	putchar(x & (1<<6) ? '1' : '0');
	putchar(x & (1<<5) ? '1' : '0');
	putchar(x & (1<<4) ? '1' : '0');
	putchar(x & (1<<3) ? '1' : '0');
	putchar(x & (1<<2) ? '1' : '0');
	putchar(x & (1<<1) ? '1' : '0');
	putchar(x & (1<<0) ? '1' : '0');
}



void print_hex(const byte *ptr, int count)
{
	int i;

	for (i=0; i<count; i++) printf("%02x", ptr[i]);
}



void print_addr(const dword ip)
{
	printf("%d.%d.%d.%d",
			(ip >> 24) & 255,
			(ip >> 16) & 255,
			(ip >> 8) & 255,
			(ip & 255)
		);
}



#define ETH_PKT_TYPE_IP 0x0800
__inline word eth_pkt_type(const byte *pkt)
{
	return ntohs( ((word*)pkt)[6] );
}

__inline byte *eth_pkt_strip(byte *pkt)
{
	return pkt+14;
}


#define IP_PROTO_ICMP 1 /* Internet Control Message Protocol */
#define IP_PROTO_IGMP 2 /* Internet Group Management Protocol */
#define IP_PROTO_IPIP 4 /* IP in IP encapsulation */
#define IP_PROTO_TCP 6  /* Transmission Control Protocol */
#define IP_PROTO_UDP 17 /* User Datagram Protocol */
__inline word ip_pkt_len(const byte *pkt)
{
	return ntohs( ((word*)pkt)[1] );
}

__inline u_char ip_pkt_proto(const byte *pkt)
{
	return pkt[9];
}

__inline dword ip_pkt_srcip(const byte *pkt)
{
	return ntohl( ((dword*)(pkt+12))[0] );
}

__inline dword ip_pkt_destip(const byte *pkt)
{
	return ntohl( ((dword*)(pkt+16))[0] );
}

__inline u_char *ip_pkt_strip(u_char *pkt)
{
	return pkt+20;
}



__inline word tcp_pkt_srcport(const byte *pkt)
{
	return ntohs( ((word*)pkt)[0] );
}

__inline word tcp_pkt_destport(const byte *pkt)
{
	return ntohs( ((word*)pkt)[1] );
}

__inline word tcp_pkt_seq(const byte *pkt)
{
	return ntohl( ((dword *)pkt)[1] );
}

__inline word tcp_pkt_ack(const byte *pkt)
{
	return ntohl( ((dword *)pkt)[2] );
}

__inline u_char tcp_flags(const byte *pkt)
{
	return pkt[13] & 0x3F;
}

#define TCP_URG (1<<5)
#define TCP_ACK (1<<4)
#define TCP_PSH (1<<3)
#define TCP_RST (1<<2)
#define TCP_SYN (1<<1)
#define TCP_FIN (1<<0)



void handle_pkt(u_char *user, const struct pcap_pkthdr *pheader, u_char *pdata)
{
	u_char *ptr;
	word pkt_type, pkt_len;
	u_char pkt_proto;
	dword pkt_srcip, pkt_destip;
	port_record *p;
	i64 one = {0,1}, inc_data = {0, (dword)(pheader->len)};

	/* update total packet and data counters */
	add_64bit(&num_packets, one);
	add_64bit(&total_data, inc_data);

	/* REDUNDANT
	if (strcmp(user, "abc") != 0) {
		printf("user = %s (wrong!)\n", user);
		return;
	}
	*/

	if (!is_ether)
	{
		ptr = pdata;
		while (!ptr[0]) ptr++;

		ptr -= 12;
	}

	pkt_type = eth_pkt_type(pdata);

	if (pkt_type != ETH_PKT_TYPE_IP) {
		if (want_packetdump)
		{
			putchar('(');
			switch (pkt_type)
			{
				case 0x0806: printf("ARP"); break;
				case 0x809b: printf("AppleTalk"); break;
				case 0x8035: printf("RARP"); break;
				case 0x814c: printf("SNMP"); break;
				case 0x86dd: printf("IPv6"); break;
				case 0x880b: printf("PPP"); break;
				default: printf("unknown");
			}
			printf(" packet -- ignored)\n");
			fflush(stdout);
		}
		return;
	}

	ptr = eth_pkt_strip(pdata);

	if (want_packetdump) printf("(IP) ");

	pkt_proto = ip_pkt_proto(ptr);
	pkt_srcip = ip_pkt_srcip(ptr);
	pkt_destip = ip_pkt_destip(ptr);
	pkt_len = ip_pkt_len(ptr);

	if (want_packetdump)
	{
		printf(proto_name_short(pkt_proto));
		putchar(' ');
		print_addr(pkt_srcip);
		printf("->");
		print_addr(pkt_destip);
	}

	/* start meddling with db */
	pthread_mutex_lock(&db_mutex);

	/* accounting for protocols */
	if (pkt_srcip == local_ip)
		proto_transfer_out(pkt_proto, pkt_len); else
	if (pkt_destip == local_ip)
		proto_transfer_in(pkt_proto, pkt_len);

	/* accounting for hosts */
	host_transfer_out(pkt_srcip, pkt_len);
	host_transfer_in(pkt_destip, pkt_len);

	/* accounting for ports */
	if (pkt_proto == IP_PROTO_TCP)
	{
		u_char *tcp_ptr = ip_pkt_strip(ptr);
		word	srcport = tcp_pkt_srcport(tcp_ptr),
			destport = tcp_pkt_destport(tcp_ptr);
		u_char flags = tcp_flags(tcp_ptr);

		if (pkt_destip == local_ip)
		{
			if ((flags & TCP_SYN) && !(flags ^ TCP_SYN))
				p = port_from_num(destport);
			else
				p = port_find(destport);
			if (p) port_update_in(p, pkt_len);
		}
	       	else if (pkt_srcip == local_ip)
		{
			p = port_find(srcport);
			if (p) port_update_out(p, pkt_len);
		}	

		if (want_packetdump) {
			printf(" %d:%d (", srcport,destport);
		
			if (flags & TCP_URG) printf("URG ");
			if (flags & TCP_ACK) printf("ACK ");
			if (flags & TCP_PSH) printf("PSH ");
			if (flags & TCP_RST) printf("RST ");
			if (flags & TCP_SYN) printf("SYN ");
			if (flags & TCP_FIN) printf("FIN ");
			if (flags) putchar(8);
			printf(") ");

			printf("seq=%u", tcp_pkt_seq(tcp_ptr));
			if (flags & TCP_ACK)
				printf(", ack=%u", tcp_pkt_ack(tcp_ptr));
		}
	}

	/* stop meddling with db */
	pthread_mutex_unlock(&db_mutex);

	/* update graph */
	pthread_mutex_lock(&graph_mutex);
	if (pkt_srcip == local_ip) graph_add_out(pkt_len); else
	if (pkt_destip == local_ip) graph_add_in(pkt_len);
	pthread_mutex_unlock(&graph_mutex);

	if (want_packetdump)
	{
		printf(" size:%d\n", pkt_len);
		fflush(stdout);
	}
}



void init_db(void)
{
	host_db_init();
	proto_db_init();
	port_db_init();
}

void free_db(void)
{
	host_db_free();
	proto_db_free();
	port_db_init();
}

void save_db(char *filename)
{
	time_t uptime = t_already + (time(NULL) - t_start);
	FILE *fp = fopen(filename, "w");
	if (!fp) {
		fprintf(stderr,"save_db(): Can't open file %s to dump db.\n", filename);
		exit(101);
	}

	fprintf(fp,"%u\n%u\n%u\n%u\n%u\n", (dword)uptime, num_packets.hi, num_packets.lo,
			total_data.hi, total_data.lo);

	host_db_dump(fp);
	proto_db_dump(fp);
	port_db_dump(fp);

	fclose(fp);
}

void load_db(char *filename)
{
	FILE *fp = fopen(filename, "r");
	char buf[1024];
	int i,j;
	
	if (!fp) {
		printf("Can't load db from %s, starting from scratch.\n", filename);
		return;
	}

#define getline if (!fgets(buf,1024,fp)) { fprintf(stderr, "load_db() failed: invalid data " \
		"in %s\n", filename); exit(101); }
	
#define atodw(buf) (dword)strtoul(buf,NULL,10)
	
#define parse(delimiter) { l += k+1;		\
		k = 0;				\
		while (buf[l+k] != delimiter)	\
		{				\
			buf2[k] = buf[l+k];	\
			k++;			\
		}				\
		buf2[k] = 0; }

	/* time */
	getline; t_already = atodw(buf);

	/* totals */
	getline; num_packets.hi=atodw(buf);
	getline; num_packets.lo=atodw(buf);
	getline; total_data.hi=atodw(buf);
	getline; total_data.lo=atodw(buf);

	/* hosts */
	getline;
	j = atodw(buf);
	for (i=0; i<j; i++)
	{
		char buf2[1024];
		int k,l;
		host_record *h;

		getline;
		
		/* field 1: ip */
		l=-1; k=0;
		parse('|');

		h = host_make(atodw(buf2));

		/* field 2: domain name */
		parse('|');
		if (strlen(buf2) == 0)
			h->hostname = NULL;
		else h->hostname = copy_str(buf2);

		/* field 3: data in pow32 */
		parse(':');
		h->data_in.hi = atodw(buf2);	

		/* field 4: data in */
		parse('|');
		h->data_in.lo = atodw(buf2);	

		/* field 5: data out pow32 */
		parse(':');
		h->data_out.hi = atodw(buf2);	

		/* field 6: data out */
		parse('\n');
		h->data_out.lo = atodw(buf2);	

		if (!db_add(&host_db, h)) freakout();
	}	
	printf("Loaded hosts.\n"); fflush(stdout);

	/* protocols */
	getline;
	j = atodw(buf);
	for (i=0; i<j; i++)
	{
		char buf2[1024];
		int k,l;
		proto_record *p;
		
		getline;

		/* field 1: protocol number */
		l=-1; k=0;
		parse('|');
		p = proto_make(atodw(buf2));
	
		/* field 2: data in pow32 */
		parse(':');
		p->data_in.hi = atodw(buf2);	

		/* field 3: data in */
		parse('|');
		p->data_in.lo = atodw(buf2);

		/* field 4: data out pow32 */
		parse(':');
		p->data_out.hi = atodw(buf2);	

		/* field 5: data out */
		parse('\n');
		p->data_out.lo = atodw(buf2);	

		if (!db_add(&proto_db, p)) freakout();
	}	
	printf("Loaded protocols.\n"); fflush(stdout);

	/* tcp ports */
	getline;
	j = atodw(buf);
	for (i=0; i<j; i++)
	{
		char buf2[1024];
		int k,l;
		port_record *p;
		
		getline;
		
		/* field 1: port number */
		l=-1; k=0;
		parse('|');

		p = port_make(atodw(buf2));
	
		/* field 2: data in pow32 */
		parse(':');
		p->data_in.hi = atodw(buf2);	

		/* field 3: data in */
		parse('|');
		p->data_in.lo = atodw(buf2);

		/* field 4: data out pow32 */
		parse(':');
		p->data_out.hi = atodw(buf2);	

		/* field 5: data out */
		parse('\n');
		p->data_out.lo = atodw(buf2);	

		if (!db_add(&port_db, p)) freakout();
	}

	printf("Loaded ports.\n"); fflush(stdout);

#undef getline
#undef atodw
#undef parse
	
	fclose(fp);
}



pcap_t *pcap;

void term_signal(int signum)
{
	fprintf(stderr,"\nCaught termination signal, dumping db...");
	fflush(stderr);

	pthread_mutex_lock(&db_mutex);
	save_db(DBFILE);
	pthread_mutex_unlock(&db_mutex);

	www_shutdown();
	pthread_kill(graph_thread, SIGTERM);
	pthread_kill(www_thread, SIGTERM);
	pthread_kill(dns_thread, SIGTERM);
	pthread_mutex_destroy(&graph_mutex);
	pthread_mutex_destroy(&db_mutex);

	free_db();
	pcap_close(pcap);

	fprintf(stderr, "done!\n");
	exit(EXIT_SUCCESS);
}



dword get_local_ip(char *inter)
{
	int tmp = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	struct ifreq ifr;
	struct sockaddr sa;

	strcpy(ifr.ifr_name, inter);
	ifr.ifr_addr.sa_family = AF_INET;
	if (ioctl(tmp, SIOCGIFADDR, &ifr) != 0)
	{
		fprintf(stderr, "Can't get own IP address.\n");
		exit(101);
	}
	close(tmp);
	sa = ifr.ifr_addr;

/* linux/socket.h:
 * struct sockaddr {
 *	sa_family_t     sa_family;      * address family, AF_xxx       
 *	char            sa_data[14];    * 14 bytes of protocol address 
 */

/* linux/in.h:
 * struct sockaddr_in {
 * sa_family_t           sin_family;     * Address family
 *	unsigned short int    sin_port;       * Port number
 *	struct in_addr        sin_addr;       * Internet address
 * * Pad to size of `struct sockaddr'.
 *
 * * Internet address.
 * struct in_addr {
 *	__u32   s_addr;
 */

	return ntohl( ((struct sockaddr_in*)&sa)->sin_addr.s_addr );
}



int main(int argc, char **argv)
{
	char *netdev, err[PCAP_ERRBUF_SIZE];
	struct pcap_stat ps;
	
	printf(PACKAGE " v" VERSION " using libpcap v%d.%d\n", PCAP_VERSION_MAJOR,
			PCAP_VERSION_MINOR);

	/* handle commandline arguments */
	if (argc < 2)
	{
		netdev = pcap_lookupdev(err);
		
		printf("usage: %s <interface> [-q]\n"
			"The interface (default: %s) specifies which interface to listen on.\n"
			"The -q argument (optional) enforces \"quiet\" mode where packet dumps\n"
			"will not be written to stdout (reduces CPU usage).\n\n",
			argv[0], netdev);

		return 101;
	} else {
		netdev = argv[1];
		if (argc == 3)
		{
			if (strcmp(argv[2],"-q") == 0)
				want_packetdump = 0;
			else {
				printf("Invalid argument: %s\n\n", argv[2]);
				return 101;
			}
		}
	}

	if (strcmp(netdev,"lo") == 0) is_ether = 0;

	local_ip = get_local_ip(netdev);
	printf("Sniffing on device %s, local IP is ", netdev);
	print_addr(local_ip);
	putchar('\n');

	pcap = pcap_open_live(netdev, 1024, 1, 0, err);
	t_start = time(NULL);

	if (!pcap) {
		fprintf(stderr,"pcap_open_live() failed: %s\n", err);
		return 101;
	}

	graph_init();
	init_db();
	load_db(DBFILE);

	if (host_db.used == 0)
		db_add(&host_db, host_make(local_ip));
	else {
		host_record *h = db_get(host_db, 0);
		if (h->ip_addr != local_ip) {
			fprintf(stderr,"Invalid database: first record must be local machine.\n");
			return 101;
		}
	}

	/* the only way out */
	signal(SIGTERM, term_signal);
	signal(SIGINT, term_signal);

	/* fire up threads */
	if (pthread_mutex_init(&db_mutex, NULL) == -1)
	{
		fprintf(stderr,"Could not create db mutex.\n");
		return 101;
	}
	if (pthread_create(&dns_thread, NULL, (void*)&dns_main, NULL) == -1)
	{
		fprintf(stderr,"Could not create DNS thread.\n");
		return 101;
	}
	if (pthread_create(&www_thread, NULL, (void*)&www_main, NULL) == -1)
	{
		fprintf(stderr,"Could not create WWW thread.\n");
		return 101;
	}
	if (pthread_mutex_init(&graph_mutex, NULL) == -1)
	{
		fprintf(stderr,"Could not create graph mutex.\n");
		return 101;
	}
	if (pthread_create(&graph_thread, NULL, (void*)&graph_main, NULL) == -1)
	{
		fprintf(stderr,"Could not create graph thread.\n");
		return 101;
	}

	printf("Entering main loop...\n");
	while (1)
	{
		/* capture ten packets */
		pcap_loop(pcap, 10, (pcap_handler)handle_pkt, "abc");

		/* print out cap statistics */	
		pcap_stats(pcap, &ps);
		printf("Packets: received %d, dropped %d\n",
				ps.ps_recv, ps.ps_drop);

		/* commit db to disk every <x> packets */
		if ((ps.ps_recv % SAVE_EVERY_x_PACKETS) == 0)
			save_db(DBFILE);
	}

	assert(0 && "Execution should never reach this point!");
	return 0;
}

